#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys

from dragon_baseline.run_classification import main as run_classification
from dragon_baseline.run_classification_multi_label import \
    main as run_classification_multi_label
from dragon_baseline.run_ner import main as run_ner
from dragon_baseline.run_ner_multi_label_binary import \
    main as run_ner_multi_label_binary

if __name__ == '__main__':
    if len(sys.argv) < 2:
        raise ValueError("First argument must specify the problem type")

    problem_type = sys.argv.pop(1)
    if problem_type == "ner":
        run_ner()
    elif problem_type == "multi_label_ner":
        run_ner_multi_label_binary()
    elif problem_type == "classification":
        run_classification()
    elif problem_type == "multi_label_classification":
        run_classification_multi_label()
    else:
        raise ValueError(f"Unexpected problem type: {problem_type}.")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pandas as pd
import torch
from evalutils import ClassificationAlgorithm
from sklearn.base import TransformerMixin


class ProblemType(Enum):
    """Problem type of the task"""
    SINGLE_LABEL_REGRESSION = "single_label_regression"
    MULTI_LABEL_REGRESSION = "multi_label_regression"
    SINGLE_LABEL_BINARY_CLASSIFICATION = "single_label_binary_classification"
    MULTI_LABEL_BINARY_CLASSIFICATION = "multi_label_binary_classification"
    SINGLE_LABEL_MULTI_CLASS_CLASSIFICATION = "single_label_multi_class_classification"
    MULTI_LABEL_MULTI_CLASS_CLASSIFICATION = "multi_label_multi_class_classification"
    SINGLE_LABEL_NER = "named_entity_recognition"
    MULTI_LABEL_BINARY_NER = "multi_label_binary_named_entity_recognition"


def string_to_ProblemType(type_str):
    """Convert a string to a corresponding Enum member if possible."""
    for problem_type in ProblemType:
        if problem_type.value == type_str:
            return problem_type

    raise ValueError(f"{type_str} does not correspond to a valid ProblemType")


@dataclass
class LabelDetails:
    """
    Details about a label.
    The `None` values are set after calling `analyze()`.
    """
    problem_type: str
    label_name: str
    prediction_name: str
    values: List[Union[str, int, float]] = None
    mean: float = None
    std: float = None
    min: float = None
    max: float = None
    skew: float = None

    @classmethod
    def from_label_name(cls, name: str):
        """Create a `LabelDetails` object from a label name as {problem_type}_target."""
        if not name.endswith("_target"):
            raise ValueError(f"Expected label name to end with '_target', but got '{name}'")
        prediction_name = name[:-len("_target")]
        problem_type = string_to_ProblemType(prediction_name)

        return cls(
            problem_type=problem_type,
            label_name=name,
            prediction_name=prediction_name,
        )


@dataclass
class TaskDetails:
    """Details about the task."""
    version: str  # version of the task details format
    jobid: int  # unique identifier for this job
    task_name: str  # name of the task
    input_name: str  # name of algorithm input ("text" for a single input string, "text_parts" for a list of strings)
    target: LabelDetails  # details about the target
    recommended_truncation_side: str  # "left" or "right"

    @classmethod
    def from_json(cls, path: Path):
        with open(path) as f:
            task_details = json.load(f)
        return cls(
            version=task_details["version"],
            jobid=task_details["jobid"],
            task_name=task_details["task_name"],
            input_name=task_details["input_name"],
            target=LabelDetails.from_label_name(task_details["label_name"]),
            recommended_truncation_side=task_details["recommended_truncation_side"],
        )


class NLPAlgorithm(ClassificationAlgorithm):
    def __init__(self, **kwargs):
        """
        The base class for NLP algorithms. Sets the environment and controls
        the flow of the processing once `process` is called.

        Parameters
        ----------
        input_path
            The path in the container where the ground truth will be loaded
            from. Default: `/input`
        output_path
            The path in the container where the output images will be written.
            Default: `/output/images`
        """

        super().__init__(index_key="input_json", **kwargs)

        # defaults
        self.df_train: pd.DataFrame = None  # loaded in self.load()
        self.df_val: pd.DataFrame = None  # loaded in self.load()
        self.df_test: pd.DataFrame = None  # loaded in self.load()
        self.task: TaskDetails = None  # loaded in self.load()
        self.label_scalers: Dict[str, TransformerMixin] = {}  # set in self.scale_labels()
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        # paths
        self.dataset_train_path = self._input_path / "nlp-training-dataset.json"
        self.dataset_val_path = self._input_path / "nlp-validation-dataset.json"
        self.dataset_test_path = self._input_path / "nlp-test-dataset.json"
        self.task_details_path = self._input_path / "nlp-task-configuration.json"
        self.test_predictions_path = self._output_path / "nlp-predictions-dataset.json"

    def process(self):
        self.load()
        self.validate()
        self.analyze()
        self.preprocess()
        self.train()
        predictions = self.predict(df=self.df_test)
        self.save(predictions)
        self.verify_predictions()

    def load_dataset(self, path: Path) -> pd.DataFrame:
        """Load a dataset."""
        df = pd.read_json(path, dtype={"uid": str})

        # cast and validate algorithm inputs
        if not "uid" in df.columns:
            raise ValueError("Expected column 'uid' in input data")
        if self.task.input_name == "text":
            df[self.task.input_name] = df[self.task.input_name].astype(str)
        elif self.task.input_name == "text_parts":
            df[self.task.input_name] = df[self.task.input_name].apply(lambda x: [str(value) for value in x])
        else:
            raise ValueError(f"Unexpected input name '{self.task.input_name}'")

        # cast and validate labels
        target = self.task.target
        if target.label_name in df.columns:
            if target.problem_type == ProblemType.SINGLE_LABEL_REGRESSION:
                df[target.label_name] = df[target.label_name].astype(float)
            elif target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
                df[target.label_name] = df[target.label_name].apply(lambda x: [float(value) if value is not None else np.nan for value in x])
            elif target.problem_type == ProblemType.SINGLE_LABEL_BINARY_CLASSIFICATION:
                df[target.label_name] = df[target.label_name].astype(int)
                if not all(value in [0, 1] for value in df[target.label_name]):
                    raise ValueError(f"Expected values in column '{target.label_name}' to be 0 or 1")
            elif target.problem_type == ProblemType.MULTI_LABEL_BINARY_CLASSIFICATION:
                df[target.label_name] = df[target.label_name].apply(lambda x: [int(value) for value in x])
                if not all(value in [0, 1] for value in df[target.label_name].explode()):
                    raise ValueError(f"Expected values in column '{target.label_name}' to be 0 or 1")
            elif target.problem_type == ProblemType.SINGLE_LABEL_MULTI_CLASS_CLASSIFICATION:
                df[target.label_name] = df[target.label_name].astype(str)
            elif target.problem_type == ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION:
                df[target.label_name] = df[target.label_name].apply(lambda x: [str(value) for value in x])
            elif target.problem_type == ProblemType.SINGLE_LABEL_NER:
                df[target.label_name] = df[target.label_name].apply(lambda x: [str(value) for value in x])
            elif target.problem_type == ProblemType.MULTI_LABEL_BINARY_NER:
                df[target.label_name] = df[target.label_name].apply(lambda x: [[bool(val) for val in value] for value in x])
            else:
                raise ValueError(f"Unexpected problem type '{target.problem_type}'")

        return df

    def load(self):
        """Load the input data."""
        self.task = TaskDetails.from_json(self.task_details_path)
        self.df_train = self.load_dataset(self.dataset_train_path)
        self.df_val = self.load_dataset(self.dataset_val_path)
        self.df_test = self.load_dataset(self.dataset_test_path)

    def validate(self):
        """Validate the input data."""
        # validate labels
        for df in [self.df_train, self.df_val]:
            if not self.task.target.label_name in df.columns:
                raise ValueError(f"Expected column '{self.task.target.label_name}' in train and validation data")

    def analyze(self):
        """
        Analyze the data.
        Collect unique labels in the training dataset.
        For regression tasks, characterize the distribution of the labels.
        """
        values = self.df_train[self.task.target.label_name].explode().explode()
        self.task.target.values = sorted(values[~values.isna()].unique().tolist())

        if self.task.target.problem_type in [ProblemType.SINGLE_LABEL_REGRESSION, ProblemType.MULTI_LABEL_REGRESSION]:
            self.task.target.mean = self.df_train[self.task.target.label_name].explode().mean()
            self.task.target.std = self.df_train[self.task.target.label_name].explode().std()
            self.task.target.min = self.df_train[self.task.target.label_name].explode().min()
            self.task.target.max = self.df_train[self.task.target.label_name].explode().max()
            self.task.target.skew = self.df_train[self.task.target.label_name].explode().skew()

    @abstractmethod
    def preprocess(self):
        """Preprocess the data."""
        raise NotImplementedError()

    @abstractmethod
    def train(self):
        """Train the model."""
        raise NotImplementedError()

    @abstractmethod
    def predict(self, *, df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError()

    def save(self, predictions: pd.DataFrame):
        """Save the predictions."""
        self.test_predictions_path.parent.mkdir(parents=True, exist_ok=True)
        predictions.to_json(self.test_predictions_path, orient="records")

    def verify_predictions(self):
        """Verify the predictions.
        
        The predictions need to satisfy the following:
        - contain a "uid" column that matches the test data.
        - contain a column with the predictions (named after the problem_type).
        - the values in each label column must be valid.
        """
        with open(self.test_predictions_path) as f:
            predictions = json.load(f)
        df = pd.DataFrame(predictions)
        if not "uid" in df.columns:
            raise ValueError("Expected column 'uid' in predictions")
        prediction_uids = set(df["uid"].tolist())
        test_uids = set(self.df_test["uid"].tolist())
        if len(prediction_uids) != len(df):
            raise ValueError("Expected unique uids in predictions")
        if prediction_uids != test_uids:
            raise ValueError("Expected predictions to contain the same uids as the test data")
        if len(df.columns) != 2:
            raise ValueError(f"Expected 2 columns in predictions, but got {len(df.columns)}: {df.columns}")

        col = self.task.target.prediction_name
        if not col in df.columns:
            raise ValueError(f"Expected column '{col}' in predictions")
        if self.task.target.problem_type == ProblemType.SINGLE_LABEL_NER:
            df_test = self.df_test.set_index("uid")
            for _, row in df.iterrows():
                # check if all values are a list of strings
                if not isinstance(row[col], list):
                    raise ValueError(f"Expected lists in column '{col}'")
                if not all(isinstance(value, str) for value in row[col]):
                    raise ValueError(f"Expected lists of strings in column '{col}'")
                # check if the number of values matches the number of words in the report
                if len(row[col]) != len(df_test.loc[row.uid, self.task.input_name]):
                    raise ValueError(f"Expected {len(row[self.task.input_name])} values in column '{col}'")
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_NER:
            df_test = self.df_test.set_index("uid")
            for _, row in df.iterrows():
                # check if all values are a list of lists of floats
                if not isinstance(row[col], list):
                    raise ValueError(f"Expected lists in column '{col}'")
                if not all(isinstance(value, list) for value in row[col]):
                    raise ValueError(f"Expected list of lists in column '{col}'")
                if not all(isinstance(value, float) for values in row[col] for value in values):
                    raise ValueError(f"Expected list of lists of floats in column '{col}'")
                # check if the number of values matches the number of words in the report
                if len(row[col]) != len(df_test.loc[row.uid, self.task.input_name]):
                    raise ValueError(f"Expected {len(row[self.task.input_name])} values in column '{col}'")
        elif self.task.target.problem_type == ProblemType.SINGLE_LABEL_BINARY_CLASSIFICATION:
            # check if all values are a float
            if not all(isinstance(value, float) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a float")
            # check if all values are in [0, 1]
            if not all(0 <= value <= 1 for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be in [0, 1]")
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_CLASSIFICATION:
            # check if all values are a list of floats
            if not all(isinstance(value, list) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a list")
            if not all(isinstance(value, float) for value in df[col].explode()):
                raise ValueError(f"Expected values in column '{col}' to be a list of floats")
            # check if all values are in [0, 1]
            if not all(0 <= value <= 1 for value in df[col].explode()):
                raise ValueError(f"Expected values in column '{col}' to be in [0, 1]")
        elif self.task.target.problem_type == ProblemType.SINGLE_LABEL_MULTI_CLASS_CLASSIFICATION:
            # check if all values are string
            if not all(isinstance(value, str) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a string")
            # check if all values are in the list of possible values
            if not all(value in self.task.target.values for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be in {self.task.target.values}")
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION:
            # check if all values are a list of strings
            if not all(isinstance(value, list) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a list")
            if not all(isinstance(value, str) for value in df[col].explode()):
                raise ValueError(f"Expected values in column '{col}' to be a list of strings")
            # check if all values are in the list of possible values
            if not all(value in self.task.target.values for value in df[col].explode()):
                raise ValueError(f"Expected values in column '{col}' to be in {self.task.target.values}")
        elif self.task.target.problem_type == ProblemType.SINGLE_LABEL_REGRESSION:
            # check if all values are a float
            if not all(isinstance(value, float) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a float")
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
            # check if all values are a list of floats
            if not all(isinstance(value, list) for value in df[col]):
                raise ValueError(f"Expected values in column '{col}' to be a list")
            if not all(isinstance(value, float) for value in df[col].explode()):
                raise ValueError(f"Expected values in column '{col}' to be a list of floats")
        else:
            raise ValueError(f"Unexpected problem type '{self.task.target.problem_type}'")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
from pathlib import Path
from typing import Iterable, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from transformers import (AutoConfig, AutoModel,
                          AutoModelForSequenceClassification, PretrainedConfig,
                          PreTrainedModel)
from transformers.modeling_outputs import SequenceClassifierOutput


class AutoModelForMultiHeadSequenceClassificationConfig(PretrainedConfig):
    model_type="AutoModelForMultiHeadSequenceClassification"

    def __init__(self, pretrained_model_name_or_path: Union[Path, str] = "bert-base-multilingual-cased", num_classes_per_label: Iterable[int] = (5, 7), **kwargs):
        super().__init__(num_labels=sum(num_classes_per_label), **kwargs)
        self.num_classes_per_label = num_classes_per_label
        self.pretrained_model_name_or_path = pretrained_model_name_or_path


class AutoModelForMultiHeadSequenceClassification(PreTrainedModel):
    config_class = AutoModelForMultiHeadSequenceClassificationConfig

    """
    This class is a wrapper around the AutoModelForSequenceClassification class
    that allows for multi-head classification with two sets of labels.
    """

    def __init__(self, config: AutoModelForMultiHeadSequenceClassificationConfig, *args, **kwargs):
        super().__init__(config)

        # set up pretrained model
        self.model = AutoModelForSequenceClassification.from_pretrained(
            config.pretrained_model_name_or_path,
            num_labels=config.num_labels,
            ignore_mismatched_sizes=True,
            *args, **kwargs
        )

    # adapted from https://github.com/huggingface/transformers/blob/faed2ca46fb163082d154aa234fd5d30682d6bf1/src/transformers/models/bert/modeling_bert.py#L1541
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], SequenceClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, 2)`, *optional*):
            Labels for computing the sequence classification loss.
        """
        return_dict = return_dict if return_dict is not None else self.model.config.use_return_dict

        # Create a dictionary of all potential arguments
        kwargs = {
            'attention_mask': attention_mask,
            'token_type_ids': token_type_ids,
            'position_ids': position_ids,
            'head_mask': head_mask,
            'inputs_embeds': inputs_embeds,
            'output_attentions': output_attentions,
            'output_hidden_states': output_hidden_states,
            'return_dict': return_dict,
        }

        # Get the list of parameter names accepted by the model method
        accepted_params = inspect.signature(self.model.forward).parameters.keys()

        # Filter the arguments to only those accepted by the model method
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in accepted_params}
        outputs: SequenceClassifierOutput = self.model(
            input_ids,
            **filtered_kwargs,
        )

        # split logits for each head
        logits = outputs.logits
        num_heads = len(self.config.num_classes_per_label)
        cumulative_indices = [0] + list(np.cumsum(self.config.num_classes_per_label))
        logits_per_head = [logits[:, cumulative_indices[i]:cumulative_indices[i+1]] for i in range(num_heads)]

        loss_fct = nn.CrossEntropyLoss()

        loss = None
        if labels is not None:
            loss = 0
            for i in range(num_heads):
                loss += loss_fct(logits_per_head[i].float(), labels[:, i].long())
            loss = loss / self.config.num_labels

        if not return_dict:
            output = logits_per_head + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits_per_head,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


# register your config and your model
AutoConfig.register("AutoModelForMultiHeadSequenceClassification", AutoModelForMultiHeadSequenceClassificationConfig)
AutoModel.register(AutoModelForMultiHeadSequenceClassificationConfig, AutoModelForMultiHeadSequenceClassification)


if __name__ == "__main__":
    # debug
    config = AutoModelForMultiHeadSequenceClassificationConfig(pretrained_model_name_or_path="bert-base-multilingual-cased", num_classes_per_label=[2, 3, 4])
    model = AutoModelForMultiHeadSequenceClassification(config)
    print(model)
    model.save_pretrained("test-workdir/model_save_load_test")
    model_reloaded = AutoModelForMultiHeadSequenceClassification.from_pretrained("test-workdir/model_save_load_test")
    print(model_reloaded)
    print("Model successfully saved and reloaded.")



#!/usr/bin/env python
# coding=utf-8
# Copyright 2020 The HuggingFace Team All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Fine-tuning the library models for token classification.
"""
# You can also adapt this script on your own token classification task and datasets. Pointers for this are left as
# comments.

import logging
import os
import sys
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import datasets
import numpy as np
import transformers
from datasets import load_dataset, load_from_disk
from scipy.special import expit
from sklearn.metrics import roc_auc_score
from transformers import (AutoConfig, AutoTokenizer,
                          DataCollatorForTokenClassification, HfArgumentParser,
                          PreTrainedTokenizerFast, Trainer, TrainingArguments,
                          set_seed)
from transformers.trainer_utils import get_last_checkpoint
from transformers.utils import check_min_version, send_example_telemetry
from transformers.utils.versions import require_version

from dragon_baseline.architectures.ner_multi_head import (
    AutoModelForMultiHeadTokenClassification, decode_labels,
    generate_label_to_id_dict)

# Will error if the minimal version of Transformers is not installed. Remove at your own risks.
check_min_version("4.34.0.dev0")

require_version("datasets>=1.8.0", "To fix: pip install -r examples/pytorch/token-classification/requirements.txt")

logger = logging.getLogger(__name__)


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where do you want to store the pretrained models downloaded from huggingface.co"},
    )
    model_revision: str = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."},
    )
    token: str = field(
        default=None,
        metadata={
            "help": (
                "The token to use as HTTP bearer authorization for remote files. If not specified, will use the token "
                "generated when running `huggingface-cli login` (stored in `~/.huggingface`)."
            )
        },
    )
    use_auth_token: bool = field(
        default=None,
        metadata={
            "help": "The `use_auth_token` argument is deprecated and will be removed in v4.34. Please use `token`."
        },
    )
    trust_remote_code: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether or not to allow for custom models defined on the Hub in their own modeling files. This option"
                "should only be set to `True` for repositories you trust and in which you have read the code, as it will"
                "execute code present on the Hub on your local machine."
            )
        },
    )
    ignore_mismatched_sizes: bool = field(
        default=False,
        metadata={"help": "Will enable to load a pretrained model whose head dimensions are different."},
    )


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    task_name: Optional[str] = field(default="ner", metadata={"help": "The name of the task (ner, pos...)."})
    dataset_name: Optional[str] = field(
        default=None, metadata={"help": "The name of the dataset to use (via the datasets library)."}
    )
    dataset_config_name: Optional[str] = field(
        default=None, metadata={"help": "The configuration name of the dataset to use (via the datasets library)."}
    )
    train_file: Optional[str] = field(
        default=None, metadata={"help": "The input training data file (a csv or JSON file)."}
    )
    validation_file: Optional[str] = field(
        default=None,
        metadata={"help": "An optional input evaluation data file to evaluate on (a csv or JSON file)."},
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={"help": "An optional input test data file to predict on (a csv or JSON file)."},
    )
    remove_columns: Optional[str] = field(
        default=None,
        metadata={"help": "The columns to remove from the dataset. Multiple columns should be separated by commas."},
    )
    text_column_name: Optional[str] = field(
        default=None, metadata={"help": "The column name of text to input in the file (a csv or JSON file)."}
    )
    label_column_name: Optional[str] = field(
        default=None, metadata={"help": "The column name of label to input in the file (a csv or JSON file)."}
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached training and evaluation sets"}
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    max_seq_length: int = field(
        default=None,
        metadata={
            "help": (
                "The maximum total input sequence length after tokenization. If set, sequences longer "
                "than this will be truncated, sequences shorter will be padded."
            )
        },
    )
    truncation_side: str = field(
        default="right",
        metadata={
            "help": (
                "The direction to truncate the sequences to max_seq_length when truncation is needed. "
                "One of ('left', 'right', 'longest_first', 'only_first'). Default: 'right'"
            )
        },
    )
    pad_to_max_length: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to pad all samples to model maximum sentence length. "
                "If False, will pad the samples dynamically when batching to the maximum length in the batch. More "
                "efficient on GPU but very bad for TPU."
            )
        },
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of training examples to this "
                "value if set."
            )
        },
    )
    max_eval_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of evaluation examples to this "
                "value if set."
            )
        },
    )
    max_predict_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of prediction examples to this "
                "value if set."
            )
        },
    )
    label_all_tokens: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether to put the label for one word on all tokens of generated by that word or just on the "
                "one (in which case the other tokens will have a padding index)."
            )
        },
    )
    return_entity_level_metrics: bool = field(
        default=False,
        metadata={"help": "Whether to return all the entity levels during evaluation or just the overall ones."},
    )

    def __post_init__(self):
        if self.dataset_name is None and self.train_file is None and self.validation_file is None:
            raise ValueError("Need either a dataset name or a training/validation file.")
        else:
            if self.train_file is not None:
                extension = self.train_file.split(".")[-1]
                assert extension in ["csv", "json"], "`train_file` should be a csv or a json file."
            if self.validation_file is not None:
                extension = self.validation_file.split(".")[-1]
                assert extension in ["csv", "json"], "`validation_file` should be a csv or a json file."
        self.task_name = self.task_name.lower()


def main():
    # See all possible arguments in src/transformers/training_args.py
    # or by passing the --help flag to this script.
    # We now keep distinct sets of args, for a cleaner separation of concerns.

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, TrainingArguments))
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        # If we pass only one argument to the script and it's the path to a json file,
        # let's parse it to get our arguments.
        model_args, data_args, training_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, data_args, training_args = parser.parse_args_into_dataclasses()

    if model_args.use_auth_token is not None:
        warnings.warn("The `use_auth_token` argument is deprecated and will be removed in v4.34.", FutureWarning)
        if model_args.token is not None:
            raise ValueError("`token` and `use_auth_token` are both specified. Please set only the argument `token`.")
        model_args.token = model_args.use_auth_token

    # Sending telemetry. Tracking the example usage helps us better allocate resources to maintain them. The
    # information sent is the one passed as arguments along with your Python/PyTorch versions.
    send_example_telemetry("run_ner", model_args, data_args)

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )

    if training_args.should_log:
        # The default of training_args.log_level is passive, so we set log level at info here to have that default.
        transformers.utils.logging.set_verbosity_info()

    log_level = training_args.get_process_log_level()
    logger.setLevel(log_level)
    datasets.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()

    # Log on each process the small summary:
    logger.warning(
        f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
        + f"distributed training: {training_args.parallel_mode.value == 'distributed'}, 16-bits training: {training_args.fp16}"
    )
    logger.info(f"Training/evaluation parameters {training_args}")

    # Detecting last checkpoint.
    last_checkpoint = None
    if os.path.isdir(training_args.output_dir) and training_args.do_train and not training_args.overwrite_output_dir:
        last_checkpoint = get_last_checkpoint(training_args.output_dir)
        if last_checkpoint is None and len(os.listdir(training_args.output_dir)) > 0:
            raise ValueError(
                f"Output directory ({training_args.output_dir}) already exists and is not empty. "
                "Use --overwrite_output_dir to overcome."
            )
        elif last_checkpoint is not None and training_args.resume_from_checkpoint is None:
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To avoid this behavior, change "
                "the `--output_dir` or add `--overwrite_output_dir` to train from scratch."
            )

    # Set seed before initializing model.
    set_seed(training_args.seed)

    # Get the datasets: you can either provide your own CSV/JSON/TXT training and evaluation files (see below)
    # or just provide the name of one of the public datasets available on the hub at https://huggingface.co/datasets/
    # (the dataset will be downloaded automatically from the datasets Hub).
    #
    # For CSV/JSON files, this script will use the column called 'text' or the first column if no column called
    # 'text' is found. You can easily tweak this behavior (see below).
    #
    # In distributed training, the load_dataset function guarantee that only one local process can concurrently
    # download the dataset.
    if data_args.dataset_name is not None:
        # Loading dataset locally
        if Path(data_args.dataset_name).exists():
            raw_datasets = load_from_disk(data_args.dataset_name)
        # Downloading and loading a dataset from the hub.
        else:
            raw_datasets = load_dataset(
                data_args.dataset_name,
                data_args.dataset_config_name,
                cache_dir=model_args.cache_dir,
                use_auth_token=True if model_args.use_auth_token else None,
            )
    else:
        data_files = {}
        if data_args.train_file is not None:
            data_files["train"] = data_args.train_file
        if data_args.validation_file is not None:
            data_files["validation"] = data_args.validation_file
        if data_args.test_file is not None:
            data_files["test"] = data_args.test_file
        extension = data_args.train_file.split(".")[-1]
        raw_datasets = load_dataset(extension, data_files=data_files, cache_dir=model_args.cache_dir)
    # See more about loading any type of standard or custom dataset (from files, python dict, pandas DataFrame, etc) at
    # https://huggingface.co/docs/datasets/loading_datasets.html.

    if data_args.remove_columns is not None:
        for split in raw_datasets.keys():
            for column in data_args.remove_columns.split(","):
                logger.info(f"removing column {column} from split {split}")
                raw_datasets[split] = raw_datasets[split].remove_columns(column)

    if training_args.do_train:
        column_names = raw_datasets["train"].column_names
        features = raw_datasets["train"].features
    else:
        column_names = raw_datasets["validation"].column_names
        features = raw_datasets["validation"].features

    if data_args.text_column_name is not None:
        text_column_name = data_args.text_column_name
    elif "tokens" in column_names:
        text_column_name = "tokens"
    else:
        text_column_name = column_names[0]

    if data_args.label_column_name is not None:
        label_column_name = data_args.label_column_name
    elif f"{data_args.task_name}_tags" in column_names:
        label_column_name = f"{data_args.task_name}_tags"
    else:
        label_column_name = column_names[1]

    # Note: we'll perform multi-label multi-hot binary classification, i.e., with multiple true/false labels per token.
    num_labels = len(raw_datasets["train"][label_column_name][0][0])
    label_to_id = generate_label_to_id_dict(num_bits=num_labels)

    # Load pretrained model and tokenizer
    #
    # Distributed training:
    # The .from_pretrained methods guarantee that only one local process can concurrently
    # download model & vocab.
    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_args.model_name_or_path,
        num_labels=num_labels,
        finetuning_task=data_args.task_name,
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        token=model_args.token,
        trust_remote_code=model_args.trust_remote_code,
    )

    tokenizer_name_or_path = model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path
    if config.model_type in {"bloom", "gpt2", "roberta"}:
        tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name_or_path,
            cache_dir=model_args.cache_dir,
            use_fast=True,
            revision=model_args.model_revision,
            token=model_args.token,
            trust_remote_code=model_args.trust_remote_code,
            add_prefix_space=True,
            truncation_side=data_args.truncation_side,
        )
    else:
        tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name_or_path,
            cache_dir=model_args.cache_dir,
            use_fast=True,
            revision=model_args.model_revision,
            token=model_args.token,
            trust_remote_code=model_args.trust_remote_code,
        )

    model_config = AutoModelForMultiHeadTokenClassification.config_class(
        pretrained_model_name_or_path=model_args.model_name_or_path,
        num_labels=num_labels,
        from_tf=bool(".ckpt" in model_args.model_name_or_path),
        config=config,
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        token=model_args.token,
        trust_remote_code=model_args.trust_remote_code,
        ignore_mismatched_sizes=model_args.ignore_mismatched_sizes,
    )
    model = AutoModelForMultiHeadTokenClassification(model_config)

    # Tokenizer check: this script requires a fast tokenizer.
    if not isinstance(tokenizer, PreTrainedTokenizerFast):
        raise ValueError(
            "This example script only works for models that have a fast tokenizer. Checkout the big table of models at"
            " https://huggingface.co/transformers/index.html#supported-frameworks to find the model types that meet"
            " this requirement"
        )

    # Set the correspondences label/ID inside the model config
    model.config.label2id = {l: i for i, l in enumerate(range(num_labels))}
    model.config.id2label = dict(enumerate(range(num_labels)))

    # Preprocessing the dataset
    # Padding strategy
    padding = "max_length" if data_args.pad_to_max_length else False

    # Tokenize all texts and align the labels with them.
    def tokenize_and_align_labels(examples):
        tokenized_inputs = tokenizer(
            examples[text_column_name],
            padding=padding,
            truncation=True,
            max_length=data_args.max_seq_length,
            # We use this argument because the texts in our dataset are lists of words (with a label for each word).
            is_split_into_words=True,
        )
        labels = []
        for i, label in enumerate(examples[label_column_name]):
            word_ids = tokenized_inputs.word_ids(batch_index=i)
            previous_word_idx = None
            label_ids = []
            for word_idx in word_ids:
                # Special tokens have a word id that is None. We set the label to -100 so they are automatically
                # ignored in the loss function.
                if word_idx is None:
                    label_ids.append(-100)
                # We set the label for the first token of each word.
                elif word_idx != previous_word_idx:
                    label_ids.append(label_to_id[tuple(label[word_idx])])
                # For the other tokens in a word, we set the label to either the current label or -100, depending on
                # the label_all_tokens flag.
                else:
                    if data_args.label_all_tokens:
                        label_ids.append(label_to_id[tuple(label[word_idx])])
                    else:
                        label_ids.append(-100)
                previous_word_idx = word_idx

            labels.append(label_ids)
        tokenized_inputs["labels"] = labels
        return tokenized_inputs

    if training_args.do_train:
        if "train" not in raw_datasets:
            raise ValueError("--do_train requires a train dataset")
        train_dataset = raw_datasets["train"]
        if data_args.max_train_samples is not None:
            max_train_samples = min(len(train_dataset), data_args.max_train_samples)
            train_dataset = train_dataset.select(range(max_train_samples))
        with training_args.main_process_first(desc="train dataset map pre-processing"):
            train_dataset = train_dataset.map(
                tokenize_and_align_labels,
                batched=True,
                num_proc=data_args.preprocessing_num_workers,
                load_from_cache_file=not data_args.overwrite_cache,
                desc="Running tokenizer on train dataset",
            )

    if training_args.do_eval:
        if "validation" not in raw_datasets:
            raise ValueError("--do_eval requires a validation dataset")
        eval_dataset = raw_datasets["validation"]
        if data_args.max_eval_samples is not None:
            max_eval_samples = min(len(eval_dataset), data_args.max_eval_samples)
            eval_dataset = eval_dataset.select(range(max_eval_samples))
        with training_args.main_process_first(desc="validation dataset map pre-processing"):
            eval_dataset = eval_dataset.map(
                tokenize_and_align_labels,
                batched=True,
                num_proc=data_args.preprocessing_num_workers,
                load_from_cache_file=not data_args.overwrite_cache,
                desc="Running tokenizer on validation dataset",
            )

    if training_args.do_predict:
        if "test" not in raw_datasets:
            raise ValueError("--do_predict requires a test dataset")
        predict_dataset = raw_datasets["test"]
        if data_args.max_predict_samples is not None:
            max_predict_samples = min(len(predict_dataset), data_args.max_predict_samples)
            predict_dataset = predict_dataset.select(range(max_predict_samples))
        with training_args.main_process_first(desc="prediction dataset map pre-processing"):
            predict_dataset = predict_dataset.map(
                tokenize_and_align_labels,
                batched=True,
                num_proc=data_args.preprocessing_num_workers,
                load_from_cache_file=not data_args.overwrite_cache,
                desc="Running tokenizer on prediction dataset",
            )

    # Data collator
    data_collator = DataCollatorForTokenClassification(tokenizer, pad_to_multiple_of=8 if training_args.fp16 else None)

    # Metrics
    def compute_metrics(p):
        # We need to decode the labels for each token.
        predictions, labels = p
        id2label = {i: label for label, i in label_to_id.items()}
        labels = decode_labels(labels=labels, id2label=id2label, num_labels=num_labels)
        labels = np.array(labels)

        # Apply sigmoid to the predictions to get probabilities.
        predictions = expit(predictions)

        # Remove ignored index (special tokens)
        mask = (labels != -100)
        true_predictions = predictions[mask]
        true_labels = labels[mask]

        if data_args.return_entity_level_metrics:
            raise NotImplementedError
        else:
            if np.unique(true_labels).size == 1:
                score = 0.0
            else:
                score = roc_auc_score(y_true=true_labels, y_score=true_predictions)

            return {
                "auroc": score,
            }

    # Initialize our Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset if training_args.do_train else None,
        eval_dataset=eval_dataset if training_args.do_eval else None,
        tokenizer=tokenizer,
        data_collator=data_collator,
        compute_metrics=compute_metrics,
    )

    # Training
    if training_args.do_train:
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif last_checkpoint is not None:
            checkpoint = last_checkpoint
        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        metrics = train_result.metrics
        trainer.save_model()  # Saves the tokenizer too for easy upload

        max_train_samples = (
            data_args.max_train_samples if data_args.max_train_samples is not None else len(train_dataset)
        )
        metrics["train_samples"] = min(max_train_samples, len(train_dataset))

        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

    # Evaluation
    if training_args.do_eval:
        logger.info("*** Evaluate ***")

        metrics = trainer.evaluate()

        max_eval_samples = data_args.max_eval_samples if data_args.max_eval_samples is not None else len(eval_dataset)
        metrics["eval_samples"] = min(max_eval_samples, len(eval_dataset))

        trainer.log_metrics("eval", metrics)
        trainer.save_metrics("eval", metrics)

    # Predict
    if training_args.do_predict:
        logger.info("*** Predict ***")

        predictions, labels, metrics = trainer.predict(predict_dataset, metric_key_prefix="predict")

        # Remove ignored index (special tokens) and binarize labels
        true_predictions = [
            [expit(p) for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]

        trainer.log_metrics("predict", metrics)
        trainer.save_metrics("predict", metrics)

        # Save predictions
        output_predictions_file = os.path.join(training_args.output_dir, "predictions.txt")
        if trainer.is_world_process_zero():
            with open(output_predictions_file, "w") as writer:
                for prediction in true_predictions:
                    writer.write("[" + ", ".join([("[" + ", ".join([str(p) for p in token_preds]) + "]") for token_preds in prediction]) + "]" + "\n")

    kwargs = {"finetuned_from": model_args.model_name_or_path, "tasks": "token-classification"}
    if data_args.dataset_name is not None:
        kwargs["dataset_tags"] = data_args.dataset_name
        if data_args.dataset_config_name is not None:
            kwargs["dataset_args"] = data_args.dataset_config_name
            kwargs["dataset"] = f"{data_args.dataset_name} {data_args.dataset_config_name}"
        else:
            kwargs["dataset"] = data_args.dataset_name

    if training_args.push_to_hub:
        trainer.push_to_hub(**kwargs)
    else:
        trainer.create_model_card(**kwargs)


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import subprocess
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import torch
from scipy.special import expit, softmax
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm
from transformers import (AutoModelForSequenceClassification,
                          AutoModelForTokenClassification, AutoTokenizer,
                          TokenClassificationPipeline)
from transformers.modeling_outputs import SequenceClassifierOutput

from dragon_baseline.architectures.clf_multi_head import \
    AutoModelForMultiHeadSequenceClassification
from dragon_baseline.architectures.ner_multi_head import \
    AutoModelForMultiHeadTokenClassification
from dragon_baseline.architectures.reg_multi_head import \
    AutoModelForMultiHeadSequenceRegression
from dragon_baseline.nlp_algorithm import NLPAlgorithm, ProblemType

__all__ = [
    # expose the algorithm classes
    "AutoModelForMultiHeadSequenceClassification",
    "AutoModelForMultiHeadTokenClassification",
    "AutoModelForMultiHeadSequenceRegression",
]

class CustomLogScaler(TransformerMixin, BaseEstimator):
    def __init__(self):
        self.standard_scaler = StandardScaler()

    def fit(self, X, y=None):
        # Apply log(1 + value) transformation
        log_X = np.log1p(X)

        # Fit the standard scaler
        self.standard_scaler.fit(log_X)
        return self

    def transform(self, X, y=None):
        # Apply log(1 + value) transformation
        log_X = np.log1p(X)

        # Transform using standard scaler
        return self.standard_scaler.transform(log_X)

    def inverse_transform(self, X):
        # Reverse the standard scaling
        inv_X = self.standard_scaler.inverse_transform(X)

        # Reverse the log(1 + value) transformation
        return np.expm1(inv_X)


class DragonBaseline(NLPAlgorithm):
    def __init__(self, input_path: Path = Path("/input"), output_path: Path = Path("/output"), workdir: Path = Path("/opt/app"), models_dir: Path = Path("/opt/app/models"), **kwargs):
        """
        Baseline implementation for the DRAGON Challenge (https://dragon.grand-challenge.org/).
        This baseline uses the HuggingFace Transformers library (https://huggingface.co/transformers/).

        The baseline must implement the following methods:
        - `preprocess`: preprocess the data
        - `train`: train the model
        - `predict`: predict the labels for the test data
        """
        super().__init__(input_path=input_path, output_path=output_path, **kwargs)

        # default training settings
        self.model_name = models_dir / "distilbert-base-multilingual-cased"
        self.per_device_train_batch_size = 4
        self.gradient_accumulation_steps = 2
        self.max_seq_length = 512
        self.learning_rate = 3e-5
        self.num_train_epochs = 5
        self.warmup_ratio = 0.1
        self.load_best_model_at_end = True
        self.metric_for_best_model = "loss"
        self.fp16 = True

        # paths for saving the preprocessed data and model checkpoints
        self.nlp_dataset_train_preprocessed_path = Path(workdir / "nlp-dataset-train-preprocessed.json")
        self.nlp_dataset_val_preprocessed_path = Path(workdir / "nlp-dataset-val-preprocessed.json")
        self.nlp_dataset_test_preprocessed_path = Path(workdir / "nlp-dataset-test-preprocessed.json")
        self.model_save_dir = Path(workdir / "checkpoints")

    def scale_labels(self) -> pd.DataFrame:
        """Scale the labels."""
        if self.task.target.problem_type in [ProblemType.SINGLE_LABEL_REGRESSION, ProblemType.MULTI_LABEL_REGRESSION]:
            if self.task.target.skew > 1:
                scaler = CustomLogScaler()
            else:
                scaler = StandardScaler()

            # fit the scaler on the training data
            scaler = scaler.fit(self.df_train[self.task.target.label_name].explode().values.astype(float).reshape(-1, 1))
            self.label_scalers[self.task.target.label_name] = scaler

            # scale the labels
            if self.task.target.problem_type == ProblemType.SINGLE_LABEL_REGRESSION:
                self.df_train[self.task.target.label_name] = scaler.transform(self.df_train[self.task.target.label_name].values.reshape(-1, 1))
                self.df_val[self.task.target.label_name] = scaler.transform(self.df_val[self.task.target.label_name].values.reshape(-1, 1))
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
                self.df_train[self.task.target.label_name] = self.df_train[self.task.target.label_name].apply(lambda x: np.ravel(scaler.transform(np.array(x).reshape(-1, 1))))
                self.df_val[self.task.target.label_name] = self.df_val[self.task.target.label_name].apply(lambda x: np.ravel(scaler.transform(np.array(x).reshape(-1, 1))))

    def unscale_predictions(self, predictions: pd.DataFrame) -> pd.DataFrame:
        """Unscale the predictions."""
        if self.task.target.problem_type in [ProblemType.SINGLE_LABEL_REGRESSION, ProblemType.MULTI_LABEL_REGRESSION]:
            scaler = self.label_scalers[self.task.target.label_name]
            if self.task.target.problem_type == ProblemType.SINGLE_LABEL_REGRESSION:
                predictions[self.task.target.prediction_name] = scaler.inverse_transform(predictions[self.task.target.prediction_name].values.reshape(-1, 1))
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
                predictions[self.task.target.prediction_name] = predictions[self.task.target.prediction_name].apply(lambda x: np.ravel(scaler.inverse_transform(np.array(x).reshape(-1, 1))))
        return predictions

    def add_dummy_test_labels(self):
        """Add dummy labels for test data. This allows to use the dataset in the huggingface pipeline."""
        if self.task.target.problem_type in [ProblemType.SINGLE_LABEL_NER, ProblemType.MULTI_LABEL_BINARY_NER]:
            dummy_label = self.df_train[self.task.target.label_name].iloc[0][0]
            self.df_test[self.task.target.label_name] = self.df_test.apply(lambda row: [dummy_label]*len(row[self.task.input_name]), axis=1)
        else:
            dummy_label = self.df_train[self.task.target.label_name].iloc[0]
            self.df_test[self.task.target.label_name] = [dummy_label]*len(self.df_test)

    def prepare_labels(self):
        """
        Prepare labels for training in the HuggingFace pipeline.

        For multi-label binary classification tasks, convert the list of 0/1 ints to list of "labelX" strings.
        For other multi-label tasks, convert the multiple values in the label column to one value per column.
        """
        if self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_CLASSIFICATION:
            for df in [self.df_train, self.df_val, self.df_test]:
                # convert list of 0/1 ints to list of "labelX" strings
                df[self.task.target.label_name] = df[self.task.target.label_name].apply(lambda labels: [
                    f"label{i}" for i, lbl in enumerate(labels) if lbl == 1
                ])
            # set the possible values (exclude NaN, which we get from `.explode()` for empty lists)
            self.task.target.values = sorted([lbl for lbl in self.df_train[self.task.target.label_name].explode().unique() if isinstance(lbl, str)])
        elif self.task.target.problem_type in [
            ProblemType.MULTI_LABEL_REGRESSION,
            ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION,
        ]:
            num_labels = len(self.df_train[self.task.target.label_name].iloc[0])
            for df in [self.df_train, self.df_val, self.df_test]:
                for i in range(num_labels):
                    df[f"{self.task.target.label_name}_{i}"] = df[self.task.target.label_name].apply(lambda x: x[i])

    def shuffle_train_data(self):
        """Shuffle the training data."""
        self.df_train = self.df_train.sample(frac=1, random_state=self.task.jobid)

    def preprocess(self):
        """Preprocess the data."""
        # scale the labels
        self.scale_labels()
        self.add_dummy_test_labels()
        self.prepare_labels()
        self.shuffle_train_data()

    def train(self):
        """Train the model."""
        # save the preprocessed data for training through command line interface of the HuggingFace library
        for path in [
            self.nlp_dataset_train_preprocessed_path,
            self.nlp_dataset_val_preprocessed_path,
            self.nlp_dataset_test_preprocessed_path,
        ]:
            path.parent.mkdir(parents=True, exist_ok=True)
        self.df_train.to_json(self.nlp_dataset_train_preprocessed_path, orient="records")
        self.df_val.to_json(self.nlp_dataset_val_preprocessed_path, orient="records")
        self.df_test.to_json(self.nlp_dataset_test_preprocessed_path, orient="records")

        # load the tokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_name, truncation_side=self.task.recommended_truncation_side)

        # train the model
        if self.task.target.problem_type == ProblemType.SINGLE_LABEL_NER:
            trainer = "ner"
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_NER:
            trainer = "multi_label_ner"
        elif self.task.target.problem_type in [ProblemType.MULTI_LABEL_REGRESSION, ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION]:
            trainer = "multi_label_classification"
        else:
            trainer = "classification"

        cmd = [
            "python", "-m", "dragon_baseline",
            trainer,
            "--do_train",
            "--do_eval",
            "--do_predict",
            "--learning_rate", self.learning_rate,
            "--model_name_or_path", self.model_name,
            "--ignore_mismatched_sizes",
            "--num_train_epochs", self.num_train_epochs,
            "--warmup_ratio", self.warmup_ratio,
            "--max_seq_length", self.max_seq_length,
            "--truncation_side", self.task.recommended_truncation_side,
            "--load_best_model_at_end", self.load_best_model_at_end,
            "--save_strategy", "epoch",
            "--evaluation_strategy", "epoch",
            "--per_device_train_batch_size", self.per_device_train_batch_size,
            "--gradient_accumulation_steps", self.gradient_accumulation_steps,
            "--train_file", self.nlp_dataset_train_preprocessed_path,
            "--validation_file", self.nlp_dataset_val_preprocessed_path,
            "--test_file", self.nlp_dataset_test_preprocessed_path,
            "--output_dir", self.model_save_dir,
            "--overwrite_output_dir",
            "--save_total_limit", "2",
            "--seed", self.task.jobid,
            "--report_to", "none",
            "--text_column_name" + ("s" if not "ner" in trainer else ""), self.task.input_name,
            "--remove_columns", "uid",
        ]
        if self.task.target.problem_type in [
            ProblemType.MULTI_LABEL_REGRESSION,
            ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION,
        ]:
            label_names = [col for col in self.df_train.columns if col.startswith(f"{self.task.target.label_name}_")]
            cmd.extend([
                "--label_column_names", ",".join(label_names),
            ])
        else:
            cmd.extend([
                "--label_column_name", self.task.target.label_name,
            ])
        if not self.task.target.problem_type in [ProblemType.SINGLE_LABEL_NER, ProblemType.MULTI_LABEL_BINARY_NER]:
            cmd.extend([
                "--text_column_delimiter", tokenizer.sep_token,
            ])
        if self.metric_for_best_model is not None:
            cmd.extend([
                "--metric_for_best_model", str(self.metric_for_best_model),
            ])
        if self.fp16:
            cmd.append("--fp16")

        cmd = [str(arg) for arg in cmd]
        print("Training command:")
        print(" ".join(cmd))
        subprocess.check_call(cmd)

    def predict_ner(self, *, df: pd.DataFrame) -> pd.DataFrame:
        """Predict the labels for the test data.

        The pipeline below returns a list of dictionaries, one for each entity. An example is shown below:
        result = [
            {'entity_group': 'SYMPTOM', 'score': 0.25475845, 'word': 'persistent cough', 'start': 22, 'end': 38},
            {'entity_group': 'DIAGNOSIS', 'score': 0.49139544, 'word': 'likely viral infection', 'start': 39, 'end': 61},
        ]

        We convert this to a list of labels, one for each word. An example is shown below:
        prediction = [B-SYMPTOM, I-SYMPTOM, B-DIAGNOSIS, I-DIAGNOSIS, I-DIAGNOSIS]
        """
        tokenizer = AutoTokenizer.from_pretrained(self.model_save_dir, truncation_side=self.task.recommended_truncation_side)
        model = AutoModelForTokenClassification.from_pretrained(self.model_save_dir)
        classifier = TokenClassificationPipeline(
            model=model,
            tokenizer=tokenizer,
            stride=tokenizer.model_max_length // 2,
            aggregation_strategy="first",
            device=self.device,
        )

        results = []
        for _, row in tqdm(df.iterrows(), desc="Predicting", total=len(df)):
            # predict
            inputs = row[self.task.input_name]
            text = " ".join(inputs)
            result = classifier(text)

            # convert to one label per word
            start, end = 0, 0
            prediction = []
            for word in inputs:
                end = start + len(word)
                found_match = False
                for entity_group in result:
                    if entity_group["start"] <= start <= entity_group["end"]:
                        if end > entity_group["end"]:
                            # Even with `aggregation_strategy="first"` a word can span multiple entities. This happens when
                            # the word is split into multiple tokens but not merged back by the `TokenClassificationPipeline`.
                            # This happens for example with dates, e.g. "01/01/2021". We take the prediction for the first part.
                            # Feel free to implement a better solution!
                            pass

                        BI_tag = "B" if start == entity_group["start"] else "I"
                        prediction.append(BI_tag + "-" + entity_group["entity_group"])
                        found_match = True
                        break

                if not found_match:
                    prediction.append("O")
                start = end + 1

            pred = {
                self.task.target.prediction_name: prediction
            }
            results.append({"uid": row["uid"], **pred})
        return pd.DataFrame(results)

    def predict_multi_label_binary_ner(self, *, df: pd.DataFrame) -> pd.DataFrame:
        """Predict the labels for the test data."""
        # load the model and tokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_save_dir, truncation_side=self.task.recommended_truncation_side)
        model = AutoModelForMultiHeadTokenClassification.from_pretrained(self.model_save_dir).to(self.device)

        # predict
        results = []
        for _, row in tqdm(df.iterrows(), desc="Predicting", total=len(df)):
            # tokenize inputs
            inputs = row[self.task.input_name]
            tokenized_inputs = tokenizer(inputs, return_tensors="pt", truncation=True, is_split_into_words=True, padding=True).to(self.device)

            # predict
            result: SequenceClassifierOutput = model(**tokenized_inputs)
            logits: np.ndarray = result.logits.detach().cpu().numpy()

            # select the logits for the first token of each word and convert to probabilities
            predictions = []
            for batch_idx in range(len(logits)):
                word_ids = tokenized_inputs.word_ids(batch_index=batch_idx)
                previous_word_idx = None
                batch_predictions = []
                for token_idx, word_idx in enumerate(word_ids):
                    if word_idx is None:
                        # skip special tokens
                        continue

                    if word_idx == previous_word_idx:
                        # skip subwords
                        continue

                    p = expit(logits[batch_idx, token_idx])
                    batch_predictions.append(p)
                    previous_word_idx = word_idx
                predictions.append(batch_predictions)
            predictions = np.array(predictions)

            if predictions.shape[1] < len(inputs):
                # pad the predictions to the length of the inputs
                # note: feel free to implement a better solution! For example, using stride > 0 in the pipeline.
                predictions = np.pad(predictions, ((0, 0), (0, len(inputs) - predictions.shape[1]), (0, 0)), mode="constant", constant_values=0)

            # convert to labels
            expected_shape = (1, len(inputs), model.config.num_labels)
            if predictions.shape != expected_shape:
                raise ValueError(f"Expected predictions to have shape {expected_shape}, but got {predictions.shape}")

            prediction = {self.task.target.prediction_name: predictions[0]}
            results.append({"uid": row["uid"], **prediction})

        return pd.DataFrame(results)

    def predict_huggingface(self, *, df: pd.DataFrame) -> pd.DataFrame:
        """Predict the labels for the test data."""
        # load the model and tokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_save_dir, truncation_side=self.task.recommended_truncation_side)
        if self.task.target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
            model = AutoModelForMultiHeadSequenceRegression.from_pretrained(self.model_save_dir).to(self.device)
        elif self.task.target.problem_type == ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION:
            model = AutoModelForMultiHeadSequenceClassification.from_pretrained(self.model_save_dir).to(self.device)
        else:
            model = AutoModelForSequenceClassification.from_pretrained(self.model_save_dir).to(self.device)

        # predict
        results = []
        for _, row in tqdm(df.iterrows(), desc="Predicting", total=len(df)):
            # tokenize inputs
            inputs = row[self.task.input_name] if self.task.input_name == "text_parts" else [row[self.task.input_name]]
            tokenized_inputs = tokenizer(*inputs, return_tensors="pt", truncation=True).to(self.device)

            # predict
            result: SequenceClassifierOutput = model(**tokenized_inputs)

            if self.task.target.problem_type == ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION:
                logits: List[np.ndarray] = [logits.detach().cpu().numpy() for logits in result.logits]
            else:
                logits: np.ndarray = result.logits.detach().cpu().numpy()

            # convert to labels
            if self.task.target.problem_type == ProblemType.SINGLE_LABEL_REGRESSION:
                expected_shape = (1, 1)
                if logits.shape != expected_shape:
                    raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits.shape}")

                prediction = {self.task.target.prediction_name: logits[0][0]}
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_REGRESSION:
                expected_shape = (1, len(self.df_train[self.task.target.label_name].iloc[0]))
                if logits.shape != expected_shape:
                    raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits.shape}")

                prediction = {self.task.target.prediction_name: logits[0]}
            elif self.task.target.problem_type == ProblemType.SINGLE_LABEL_BINARY_CLASSIFICATION:
                expected_shape = (1, 2)
                if logits.shape != expected_shape:
                    raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits.shape}")

                # calculate sigmoid to map the logits to [0, 1]
                prediction = softmax(logits, axis=-1)[0, 1]
                prediction = {self.task.target.prediction_name: prediction}
            elif self.task.target.problem_type == ProblemType.SINGLE_LABEL_MULTI_CLASS_CLASSIFICATION:
                expected_shape = (1, len(self.task.target.values))
                if logits.shape != expected_shape:
                    raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits.shape}")

                p = model.config.id2label[np.argmax(logits[0])]
                prediction = {self.task.target.prediction_name: p}
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_CLASSIFICATION:
                expected_shape = (1, len(self.task.target.values))
                if logits.shape != expected_shape:
                    raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits.shape}")

                prediction = expit(logits)[0]  # calculate sigmoid to map the logits to [0, 1]
                prediction = {self.task.target.prediction_name: prediction}
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_MULTI_CLASS_CLASSIFICATION:
                expected_length = len(self.df_train[self.task.target.label_name].iloc[0])
                if len(logits) != expected_length:
                    raise ValueError(f"Expected logits to have length {expected_length}, but got {len(logits)}")
                label_names = [f"{self.task.target.label_name}_{i}" for i in range(len(logits))]
                for logits_, label_name in zip(logits, label_names):
                    expected_shape = (1, len(self.df_train[label_name].unique()))
                    if logits_.shape != expected_shape:
                        raise ValueError(f"Expected logits to have shape {expected_shape}, but got {logits_.shape}")

                preds = [np.argmax(p) for p in logits]
                prediction = {
                    self.task.target.prediction_name: [
                        id2label[str(p)]
                        for p, id2label in zip(preds, model.config.id2labels)
                    ]
                }
            else:
                raise ValueError(f"Unexpected problem type '{self.task.target.problem_type}'")

            results.append({"uid": row["uid"], **prediction})

        df_pred = pd.DataFrame(results)

        # scale the predictions (inverse of the normalization during preprocessing)
        df_pred = self.unscale_predictions(df_pred)

        return df_pred

    def predict(self, *, df: pd.DataFrame) -> pd.DataFrame:
        """Predict the labels for the test data."""
        with torch.no_grad():
            if self.task.target.problem_type == ProblemType.SINGLE_LABEL_NER:
                return self.predict_ner(df=df)
            elif self.task.target.problem_type == ProblemType.MULTI_LABEL_BINARY_NER:
                return self.predict_multi_label_binary_ner(df=df)
            else:
                return self.predict_huggingface(df=df)


if __name__ == "__main__":
    # Note: to debug (outside of Docker), you can set the input and output paths.
    # You probably need to change self.model_name too.
    for job_name in [
        "Task000_Example_clf-fold0",
        "Task001_Example_reg-fold0",
        "Task002_Example_mutli_reg-fold0",
        "Task003_Example_mednli-fold0",
        "Task004_Example_ner-fold0",
        "Task005_Example_multi_clf-fold0",
        "Task006_Example_binary_clf-fold0",
        "Task007_Example_multi_binary_clf-fold0",
        "Task008_Example_multi_ner-fold0",
    ]:
        DragonBaseline(
            input_path=Path(f"test-input/{job_name}"),
            output_path=Path(f"test-output/{job_name}"),
            workdir=Path(f"test-workdir/{job_name}"),
            models_dir=Path(f"test-workdir/models"),
        ).process()



#!/usr/bin/env python
# coding=utf-8
# Copyright 2020 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Finetuning the library models for text classification."""
# You can also adapt this script on your own text classification task. Pointers for this are left as comments.

import logging
import os
import random
import sys
import warnings
from dataclasses import dataclass, field
from typing import List, Optional

import datasets
import evaluate
import numpy as np
import transformers
from datasets import load_dataset
from transformers import (AutoConfig, AutoTokenizer, DataCollatorWithPadding,
                          EvalPrediction, HfArgumentParser, Trainer,
                          TrainingArguments, default_data_collator, set_seed)
from transformers.trainer_utils import get_last_checkpoint
from transformers.utils import check_min_version, send_example_telemetry
from transformers.utils.versions import require_version

from dragon_baseline.architectures.clf_multi_head import \
    AutoModelForMultiHeadSequenceClassification
from dragon_baseline.architectures.reg_multi_head import \
    AutoModelForMultiHeadSequenceRegression

# Will error if the minimal version of Transformers is not installed. Remove at your own risks.
check_min_version("4.34.0.dev0")

require_version("datasets>=1.8.0", "To fix: pip install -r examples/pytorch/text-classification/requirements.txt")


logger = logging.getLogger(__name__)


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.

    Using `HfArgumentParser` we can turn this class
    into argparse arguments to be able to specify them on
    the command line.
    """

    dataset_name: Optional[str] = field(
        default=None, metadata={"help": "The name of the dataset to use (via the datasets library)."}
    )
    dataset_config_name: Optional[str] = field(
        default=None, metadata={"help": "The configuration name of the dataset to use (via the datasets library)."}
    )
    problem_type: str = field(
        default=None,
        metadata={
            "help": "Problem type (regression, classification, multi-label regression, multi-label classification). If None, will be inferred from the dataset."
        },
    )
    text_column_names: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "The name of the text column in the input dataset or a CSV/JSON file."
                'If not specified, will use the "sentence" column for single/multi-label classifcation task.'
            )
        },
    )
    text_column_delimiter: Optional[str] = field(
        default=" ", metadata={"help": "THe delimiter to use to join text columns into a single sentence."}
    )
    train_split_name: Optional[str] = field(
        default=None,
        metadata={
            "help": 'The name of the train split in the input dataset. If not specified, will use the "train" split when do_train is enabled'
        },
    )
    validation_split_name: Optional[str] = field(
        default=None,
        metadata={
            "help": 'The name of the validation split in the input dataset. If not specified, will use the "validation" split when do_eval is enabled'
        },
    )
    test_split_name: Optional[str] = field(
        default=None,
        metadata={
            "help": 'The name of the test split in the input dataset. If not specified, will use the "test" split when do_predict is enabled'
        },
    )
    remove_splits: Optional[str] = field(
        default=None,
        metadata={"help": "The splits to remove from the dataset. Multiple splits should be separated by commas."},
    )
    remove_columns: Optional[str] = field(
        default=None,
        metadata={"help": "The columns to remove from the dataset. Multiple columns should be separated by commas."},
    )
    label_column_names: Optional[str] = field(
        default=None,
        metadata={
            "help": (
                "The names of the label columns in the input dataset or a CSV/JSON file."
                'If not specified, will use the "label1"/"label2"/etc. columns for multi-label regression/classifcation task'
            )
        },
    )
    max_seq_length: int = field(
        default=128,
        metadata={
            "help": (
                "The maximum total input sequence length after tokenization. Sequences longer "
                "than this will be truncated, sequences shorter will be padded."
            )
        },
    )
    truncation_side: str = field(
        default="right",
        metadata={
            "help": (
                "The direction to truncate the sequences to max_seq_length when truncation is needed. "
                "One of ('left', 'right', 'longest_first', 'only_first'). Default: 'right'"
            )
        },
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached preprocessed datasets or not."}
    )
    pad_to_max_length: bool = field(
        default=True,
        metadata={
            "help": (
                "Whether to pad all samples to `max_seq_length`. "
                "If False, will pad the samples dynamically when batching to the maximum length in the batch."
            )
        },
    )
    shuffle_train_dataset: bool = field(
        default=False, metadata={"help": "Whether to shuffle the train dataset or not."}
    )
    shuffle_seed: int = field(
        default=42, metadata={"help": "Random seed that will be used to shuffle the train dataset."}
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of training examples to this "
                "value if set."
            )
        },
    )
    max_eval_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of evaluation examples to this "
                "value if set."
            )
        },
    )
    max_predict_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of prediction examples to this "
                "value if set."
            )
        },
    )
    metric_name: Optional[str] = field(default=None, metadata={"help": "The metric to use for evaluation."})
    train_file: Optional[str] = field(
        default=None, metadata={"help": "A csv or a json file containing the training data."}
    )
    validation_file: Optional[str] = field(
        default=None, metadata={"help": "A csv or a json file containing the validation data."}
    )
    test_file: Optional[str] = field(default=None, metadata={"help": "A csv or a json file containing the test data."})

    def __post_init__(self):
        if self.dataset_name is None:
            if self.train_file is None or self.validation_file is None:
                raise ValueError(" training/validation file or a dataset name.")

            train_extension = self.train_file.split(".")[-1]
            assert train_extension in ["csv", "json"], "`train_file` should be a csv or a json file."
            validation_extension = self.validation_file.split(".")[-1]
            assert (
                validation_extension == train_extension
            ), "`validation_file` should have the same extension (csv or json) as `train_file`."


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where do you want to store the pretrained models downloaded from huggingface.co"},
    )
    use_fast_tokenizer: bool = field(
        default=True,
        metadata={"help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."},
    )
    model_revision: str = field(
        default="main",
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."},
    )
    token: str = field(
        default=None,
        metadata={
            "help": (
                "The token to use as HTTP bearer authorization for remote files. If not specified, will use the token "
                "generated when running `huggingface-cli login` (stored in `~/.huggingface`)."
            )
        },
    )
    use_auth_token: bool = field(
        default=None,
        metadata={
            "help": "The `use_auth_token` argument is deprecated and will be removed in v4.34. Please use `token`."
        },
    )
    trust_remote_code: bool = field(
        default=False,
        metadata={
            "help": (
                "Whether or not to allow for custom models defined on the Hub in their own modeling files. This option"
                "should only be set to `True` for repositories you trust and in which you have read the code, as it will"
                "execute code present on the Hub on your local machine."
            )
        },
    )
    ignore_mismatched_sizes: bool = field(
        default=False,
        metadata={"help": "Will enable to load a pretrained model whose head dimensions are different."},
    )


def get_label_list(raw_dataset, split="train") -> List[str]:
    """Get the list of labels from a mutli-label dataset"""

    if isinstance(raw_dataset[split]["label"][0], list):
        label_list = [label for sample in raw_dataset[split]["label"] for label in sample]
        label_list = list(set(label_list))
    else:
        label_list = raw_dataset[split].unique("label")
    # we will treat the label list as a list of string instead of int, consistent with model.config.label2id
    label_list = [str(label) for label in label_list]
    return label_list


def main():
    # See all possible arguments in src/transformers/training_args.py
    # or by passing the --help flag to this script.
    # We now keep distinct sets of args, for a cleaner separation of concerns.

    parser = HfArgumentParser((ModelArguments, DataTrainingArguments, TrainingArguments))
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        # If we pass only one argument to the script and it's the path to a json file,
        # let's parse it to get our arguments.
        model_args, data_args, training_args = parser.parse_json_file(json_file=os.path.abspath(sys.argv[1]))
    else:
        model_args, data_args, training_args = parser.parse_args_into_dataclasses()

    if model_args.use_auth_token is not None:
        warnings.warn("The `use_auth_token` argument is deprecated and will be removed in v4.34.", FutureWarning)
        if model_args.token is not None:
            raise ValueError("`token` and `use_auth_token` are both specified. Please set only the argument `token`.")
        model_args.token = model_args.use_auth_token

    # Sending telemetry. Tracking the example usage helps us better allocate resources to maintain them. The
    # information sent is the one passed as arguments along with your Python/PyTorch versions.
    send_example_telemetry("run_classification", model_args, data_args)

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )

    if training_args.should_log:
        # The default of training_args.log_level is passive, so we set log level at info here to have that default.
        transformers.utils.logging.set_verbosity_info()

    log_level = training_args.get_process_log_level()
    logger.setLevel(log_level)
    datasets.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()

    # Log on each process the small summary:
    logger.warning(
        f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
        + f"distributed training: {training_args.parallel_mode.value == 'distributed'}, 16-bits training: {training_args.fp16}"
    )
    logger.info(f"Training/evaluation parameters {training_args}")

    # Detecting last checkpoint.
    last_checkpoint = None
    if os.path.isdir(training_args.output_dir) and training_args.do_train and not training_args.overwrite_output_dir:
        last_checkpoint = get_last_checkpoint(training_args.output_dir)
        if last_checkpoint is None and len(os.listdir(training_args.output_dir)) > 0:
            raise ValueError(
                f"Output directory ({training_args.output_dir}) already exists and is not empty. "
                "Use --overwrite_output_dir to overcome."
            )
        elif last_checkpoint is not None and training_args.resume_from_checkpoint is None:
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To avoid this behavior, change "
                "the `--output_dir` or add `--overwrite_output_dir` to train from scratch."
            )

    # Set seed before initializing model.
    set_seed(training_args.seed)

    # Get the datasets: you can either provide your own CSV/JSON training and evaluation files, or specify a dataset name
    # to load from huggingface/datasets. In ether case, you can specify a the key of the column(s) containing the text and
    # the key of the column containing the label. If multiple columns are specified for the text, they will be joined togather
    # for the actual text value.
    # In distributed training, the load_dataset function guarantee that only one local process can concurrently
    # download the dataset.
    if data_args.dataset_name is not None:
        # Downloading and loading a dataset from the hub.
        raw_datasets = load_dataset(
            data_args.dataset_name,
            data_args.dataset_config_name,
            cache_dir=model_args.cache_dir,
            token=model_args.token,
        )
        # Try print some info about the dataset
        logger.info(f"Dataset loaded: {raw_datasets}")
        logger.info(raw_datasets)
    else:
        # Loading a dataset from your local files.
        # CSV/JSON training and evaluation files are needed.
        data_files = {"train": data_args.train_file, "validation": data_args.validation_file}

        # Get the test dataset: you can provide your own CSV/JSON test file
        if training_args.do_predict:
            if data_args.test_file is not None:
                train_extension = data_args.train_file.split(".")[-1]
                test_extension = data_args.test_file.split(".")[-1]
                assert (
                    test_extension == train_extension
                ), "`test_file` should have the same extension (csv or json) as `train_file`."
                data_files["test"] = data_args.test_file
            else:
                raise ValueError("Need either a dataset name or a test file for `do_predict`.")

        for key in data_files.keys():
            logger.info(f"load a local file for {key}: {data_files[key]}")

        if data_args.train_file.endswith(".csv"):
            # Loading a dataset from local csv files
            raw_datasets = load_dataset(
                "csv",
                data_files=data_files,
                cache_dir=model_args.cache_dir,
                token=model_args.token,
            )
        else:
            # Loading a dataset from local json files
            raw_datasets = load_dataset(
                "json",
                data_files=data_files,
                cache_dir=model_args.cache_dir,
                token=model_args.token,
            )

    # See more about loading any type of standard or custom dataset at
    # https://huggingface.co/docs/datasets/loading_datasets.html.

    if data_args.remove_splits is not None:
        for split in data_args.remove_splits.split(","):
            logger.info(f"removing split {split}")
            raw_datasets.pop(split)

    if data_args.train_split_name is not None:
        logger.info(f"using {data_args.train_split_name} as training set")
        raw_datasets["train"] = raw_datasets[data_args.train_split_name]
        raw_datasets.pop(data_args.train_split_name)

    if data_args.validation_split_name is not None:
        logger.info(f"using {data_args.validation_split_name} as validation set")
        raw_datasets["validation"] = raw_datasets[data_args.validation_split_name]
        raw_datasets.pop(data_args.validation_split_name)

    if data_args.test_split_name is not None:
        logger.info(f"using {data_args.test_split_name} as test set")
        raw_datasets["test"] = raw_datasets[data_args.test_split_name]
        raw_datasets.pop(data_args.test_split_name)

    if data_args.remove_columns is not None:
        for split in raw_datasets.keys():
            for column in data_args.remove_columns.split(","):
                logger.info(f"removing column {column} from split {split}")
                raw_datasets[split] = raw_datasets[split].remove_columns(column)

    if data_args.label_column_names is not None:
        logger.info(f"using {data_args.label_column_names} as label columns")
        for i, label_column_name in enumerate(data_args.label_column_names.split(",")):
            logger.info(f"using {label_column_name} as label{i+1} column")
            for key in raw_datasets.keys():
                new_label_column_name = f"label{i+1}"
                if label_column_name != new_label_column_name:
                    raw_datasets[key] = raw_datasets[key].rename_column(label_column_name, new_label_column_name)

    # Trying to have good defaults here, don't hesitate to tweak to your needs.
    if data_args.problem_type is None:
        # Infer problem_type from the high-level structure of the dataset
        if "label1" in raw_datasets["train"].features:
            # Multi-label classification or regression
            if (raw_datasets["train"].features["label1"].dtype in ["float32", "float64"]):
                data_args.problem_type = "multi_label_regression"
            else:
                data_args.problem_type = "multi_label_multi_class_classification"
        else:
            raise ValueError(
                "You are running a fine-tuning task but didn't provide label columns named 'label1'/'label2' in your CSV/JSON dataset."
            )
        logger.info(f"Setting problem type to {data_args.problem_type}.")

    # Verify dataset structure
    if data_args.problem_type in ["multi_label_regression", "multi_label_multi_class_classification"]:
        if "label1" not in raw_datasets["train"].features or "label2" not in raw_datasets["train"].features:
            raise ValueError(
                "You are running a multi-label task but didn't provide label columns named 'label1'/'label2'/etc. in your CSV/JSON dataset."
            )

    # Infer the number of labels and label mapping from the dataset
    label_names = [lbl for lbl in raw_datasets["train"].features if lbl.startswith("label")]
    label_names.sort()  # Let's sort it for determinism
    if data_args.problem_type == "multi_label_regression":
        num_labels = len(label_names)
    elif data_args.problem_type == "multi_label_multi_class_classification":
        # Trying to find the number of labels in a multi-label classification task
        # We have to deal with the scenario that labels appear in the validation/test set but not in the training set.
        # So we build the label list from the union of labels in train/val/test.
        label_lists = [raw_datasets["train"].unique(lbl) for lbl in label_names]
        for split in ["validation", "test"]:
            if split in raw_datasets:
                val_or_test_label_lists = [raw_datasets[split].unique(lbl) for lbl in label_names]
                for label_list, val_or_test_label_list in zip(label_lists, val_or_test_label_lists):
                    diff = set(val_or_test_label_list).difference(set(label_list))
                    if len(diff) > 0:
                        # labels appear in val/test but not in train, throw a warning
                        logger.warning(
                            f"Labels {diff} in {split} set but not in training set, adding them to the label list"
                        )
                        label_list.extend(list(diff))

        num_classes_per_label = [len(lbl_list) for lbl_list in label_lists]
        num_labels = sum(num_classes_per_label)
        for label_list in label_lists:
            label_list.sort()  # Let's sort it for determinism
    else:
        raise ValueError(f"Unrecognized problem type {data_args.problem_type}")

    # Load pretrained model and tokenizer
    # In distributed training, the .from_pretrained methods guarantee that only one local process can concurrently
    # download model & vocab.
    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_args.model_name_or_path,
        num_labels=num_labels,
        finetuning_task="text-classification",
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        token=model_args.token,
        trust_remote_code=model_args.trust_remote_code,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
        use_fast=model_args.use_fast_tokenizer,
        revision=model_args.model_revision,
        token=model_args.token,
        trust_remote_code=model_args.trust_remote_code,
        truncation_side=data_args.truncation_side,
    )
    if data_args.problem_type == "multi_label_multi_class_classification":
        model_config = AutoModelForMultiHeadSequenceClassification.config_class(
            pretrained_model_name_or_path=model_args.model_name_or_path,
            num_classes_per_label=num_classes_per_label,
            from_tf=bool(".ckpt" in model_args.model_name_or_path),
            config=config,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            token=model_args.token,
            trust_remote_code=model_args.trust_remote_code,
        )
        model = AutoModelForMultiHeadSequenceClassification(model_config)
    elif data_args.problem_type == "multi_label_regression":
        model_config = AutoModelForMultiHeadSequenceRegression.config_class(
            pretrained_model_name_or_path=model_args.model_name_or_path,
            num_classes_per_label=num_labels,
            from_tf=bool(".ckpt" in model_args.model_name_or_path),
            config=config,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            token=model_args.token,
            trust_remote_code=model_args.trust_remote_code,
        )
        model = AutoModelForMultiHeadSequenceRegression(model_config)
    else:
        raise ValueError(f"Unrecognized problem type {data_args.problem_type}")

    # Padding strategy
    if data_args.pad_to_max_length:
        padding = "max_length"
    else:
        # We will pad later, dynamically at batch creation, to the max sequence length in each batch
        padding = False

    label_to_ids = None
    if data_args.problem_type == "multi_label_multi_class_classification":
        label_to_ids = [
            {v: i for i, v in enumerate(label_list)}
            for label_list in label_lists
        ]

    if label_to_ids is not None:
        model.config.label2ids = label_to_ids
        model.config.id2labels = [{id: label for label, id in label2id.items()} for label2id in label_to_ids]

    if data_args.max_seq_length > tokenizer.model_max_length:
        logger.warning(
            f"The max_seq_length passed ({data_args.max_seq_length}) is larger than the maximum length for the"
            f"model ({tokenizer.model_max_length}). Using max_seq_length={tokenizer.model_max_length}."
        )
    max_seq_length = min(data_args.max_seq_length, tokenizer.model_max_length)

    def preprocess_function(examples):
        if data_args.text_column_names is not None:
            text_column_names = data_args.text_column_names.split(",")
            # join together text columns into "sentence" column
            examples["sentence"] = examples[text_column_names[0]]
            for column in text_column_names[1:]:
                for i in range(len(examples[column])):
                    examples["sentence"][i] += data_args.text_column_delimiter + examples[column][i]
        # Tokenize the texts
        result = tokenizer(examples["sentence"], padding=padding, max_length=max_seq_length, truncation=True)
        if label_to_ids is not None:
            result["label"] = np.stack(
                [
                    [(label_to_id[l] if l != -1 else -1) for l in examples[label_name]]
                    for (label_to_id, label_name) in zip(label_to_ids, label_names)
                ], axis=1
            )
        else:
            result["label"] = np.stack(
                [
                    [(l if l is not None else np.nan) for l in examples[label_name]]
                    for label_name in label_names
                ], axis=1
            )
        return result

    # Running the preprocessing pipeline on all the datasets
    with training_args.main_process_first(desc="dataset map pre-processing"):
        raw_datasets = raw_datasets.map(
            preprocess_function,
            batched=True,
            load_from_cache_file=not data_args.overwrite_cache,
            desc="Running tokenizer on dataset",
        )

    if training_args.do_train:
        if "train" not in raw_datasets:
            raise ValueError("--do_train requires a train dataset.")
        train_dataset = raw_datasets["train"]
        if data_args.shuffle_train_dataset:
            logger.info("Shuffling the training dataset")
            train_dataset = train_dataset.shuffle(seed=data_args.shuffle_seed)
        if data_args.max_train_samples is not None:
            max_train_samples = min(len(train_dataset), data_args.max_train_samples)
            train_dataset = train_dataset.select(range(max_train_samples))

    if training_args.do_eval:
        if "validation" not in raw_datasets and "validation_matched" not in raw_datasets:
            if "test" not in raw_datasets and "test_matched" not in raw_datasets:
                raise ValueError("--do_eval requires a validation or test dataset if validation is not defined.")
            else:
                logger.warning("Validation dataset not found. Falling back to test dataset for validation.")
                eval_dataset = raw_datasets["test"]
        else:
            eval_dataset = raw_datasets["validation"]

        if data_args.max_eval_samples is not None:
            max_eval_samples = min(len(eval_dataset), data_args.max_eval_samples)
            eval_dataset = eval_dataset.select(range(max_eval_samples))

    if training_args.do_predict or data_args.test_file is not None:
        if "test" not in raw_datasets:
            raise ValueError("--do_predict requires a test dataset")
        predict_dataset = raw_datasets["test"]
        # remove label column if it exists
        if data_args.max_predict_samples is not None:
            max_predict_samples = min(len(predict_dataset), data_args.max_predict_samples)
            predict_dataset = predict_dataset.select(range(max_predict_samples))

    # Log a few random samples from the training set:
    if training_args.do_train:
        for index in random.sample(range(len(train_dataset)), 3):
            logger.info(f"Sample {index} of the training set: {train_dataset[index]}.")

    if data_args.metric_name is not None:
        metric = evaluate.load(data_args.metric_name, config_name="multilabel")
        logger.info(f"Using metric {data_args.metric_name} for evaluation.")
    else:
        if data_args.problem_type == "multi_label_regression":
            metric = evaluate.load("mse", config_name="multilabel")
            logger.info(
                "Using mean squared error (mse) as regression score, you can use --metric_name to overwrite."
            )
        elif data_args.problem_type == "multi_label_multi_class_classification":
            metric = evaluate.load("accuracy")
            logger.info("Using accuracy for multi-label multi-class classification task, you can use --metric_name to overwrite.")
        else:
            raise ValueError(f"Unrecognized problem type {data_args.problem_type}")

    # You can define your custom compute_metrics function. It takes an `EvalPrediction` object (a namedtuple with a
    # predictions and label_ids field) and has to return a dictionary string to float.
    def compute_metrics(p: EvalPrediction):
        if data_args.problem_type == "multi_label_multi_class_classification":
            preds = p.predictions[0] if isinstance(p.predictions, tuple) else p.predictions
            preds = [np.argmax(logits, axis=-1) for logits in preds]
            preds = np.stack(preds, axis=1)
            preds = np.ravel(preds).astype(np.int32)
            labels = np.ravel(p.label_ids).astype(np.int32)
        elif data_args.problem_type == "multi_label_regression":
            preds = p.predictions[0] if isinstance(p.predictions, tuple) else p.predictions
            labels = p.label_ids
            mask = ~np.isnan(labels)
            preds = preds[mask]
            labels = labels[mask]
        result = metric.compute(predictions=preds, references=labels)
        if len(result) > 1:
            result["combined_score"] = np.mean(list(result.values())).item()
        return result

    # Data collator will default to DataCollatorWithPadding when the tokenizer is passed to Trainer, so we change it if
    # we already did the padding.
    if data_args.pad_to_max_length:
        data_collator = default_data_collator
    elif training_args.fp16:
        data_collator = DataCollatorWithPadding(tokenizer, pad_to_multiple_of=8)
    else:
        data_collator = None

    # Initialize our Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset if training_args.do_train else None,
        eval_dataset=eval_dataset if training_args.do_eval else None,
        compute_metrics=compute_metrics,
        tokenizer=tokenizer,
        data_collator=data_collator,
    )

    # Training
    if training_args.do_train:
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif last_checkpoint is not None:
            checkpoint = last_checkpoint
        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        metrics = train_result.metrics
        max_train_samples = (
            data_args.max_train_samples if data_args.max_train_samples is not None else len(train_dataset)
        )
        metrics["train_samples"] = min(max_train_samples, len(train_dataset))
        trainer.save_model()  # Saves the tokenizer too for easy upload
        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

    # Evaluation
    if training_args.do_eval:
        logger.info("*** Evaluate ***")
        metrics = trainer.evaluate(eval_dataset=eval_dataset)
        max_eval_samples = data_args.max_eval_samples if data_args.max_eval_samples is not None else len(eval_dataset)
        metrics["eval_samples"] = min(max_eval_samples, len(eval_dataset))
        trainer.log_metrics("eval", metrics)
        trainer.save_metrics("eval", metrics)

    if training_args.do_predict:
        logger.info("*** Predict ***")
        # Removing the `label` columns if exists because it might contains -1 and Trainer won't like that.
        if "label" in predict_dataset.features:
            predict_dataset = predict_dataset.remove_columns("label")
        predictions = trainer.predict(predict_dataset, metric_key_prefix="predict").predictions
        if data_args.problem_type == "multi_label_regression":
            if predictions.shape != (len(predict_dataset), num_labels):
                raise ValueError(
                    f"For regression, the predictions shape ({predictions.shape}) should be the number of samples x number of labels ({len(predict_dataset)}, {num_labels})."
                )
        elif data_args.problem_type == "multi_label_multi_class_classification":
            predictions = [np.argmax(logits, axis=-1) for logits in predictions]
            predictions = np.stack(predictions, axis=1)
            if predictions.shape != (len(predict_dataset), len(num_classes_per_label)):
                raise ValueError(
                    f"For multi-label multi-class classification, the predictions shape ({predictions.shape}) should be the number of samples x number of labels ({len(predict_dataset)}, {len(num_classes_per_label)})."
                )

        output_predict_file = os.path.join(training_args.output_dir, "predict_results.txt")
        if trainer.is_world_process_zero():
            with open(output_predict_file, "w") as writer:
                logger.info("***** Predict results *****")
                prediction_names = [lbl.replace("label", "prediction") for lbl in label_names]
                writer.write("index\t" + "\t".join(prediction_names) + "\n")
                for index, item in enumerate(predictions):
                    if data_args.problem_type == "multi_label_regression":
                        writer.write(f"{index}\t" + "\t".join([f"{i}" for i in item]) + "\n")
                    elif data_args.problem_type == "multi_label_multi_class_classification":
                        # recover from multi-hot encoding
                        model.config.id2labels
                        item = [
                            id2label[p]
                            for p, id2label in zip(item, model.config.id2labels)
                        ]
                        writer.write(f"{index}\t" + "\t".join(item) + "\n")
                    else:
                        raise ValueError(f"Unrecognized problem type {data_args.problem_type}")
        logger.info("Predict results saved at {}".format(output_predict_file))
    kwargs = {"finetuned_from": model_args.model_name_or_path, "tasks": "text-classification"}

    if training_args.push_to_hub:
        trainer.push_to_hub(**kwargs)
    else:
        trainer.create_model_card(**kwargs)


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
from itertools import product
from pathlib import Path
from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
from transformers import (AutoConfig, AutoModel,
                          AutoModelForTokenClassification, PretrainedConfig,
                          PreTrainedModel)
from transformers.modeling_outputs import SequenceClassifierOutput

# For documentation on sharing custom models, plaese see:
# https://huggingface.co/docs/transformers/custom_models#registering-a-model-with-custom-code-to-the-auto-classes

def generate_label_to_id_dict(num_bits):
    # Generate all combinations of True/False for the given number of bits.
    keys = product([True, False], repeat=num_bits)

    # Initialize an empty dictionary.
    label_to_id = {}

    # Populate the dictionary.
    for key in keys:
        # Convert the True/False combination to a binary string, with True as '1' and False as '0'.
        binary_str = ''.join('1' if bit else '0' for bit in key)

        # Convert the binary string to an integer.
        binary_int = int(binary_str, 2)

        # Add to the dictionary.
        label_to_id[key] = binary_int

    return label_to_id


def generate_id_to_label_dict(num_bits):
    label_to_id = generate_label_to_id_dict(num_bits)
    id_to_label = {v: k for k, v in label_to_id.items()}
    return id_to_label


def decode_labels(labels, id2label: dict, num_labels: int):
    # decode the labels to the binary vectors, while keeping -100 to ignore the loss computation
    decoded_labels = []

    # Apply the mapping to each element
    for label in labels:
        decoded_label = []
        for lbl in label:
            if lbl == -100:
                decoded_label.append([-100] * num_labels)
            else:
                if isinstance(lbl, torch.Tensor):
                    lbl = lbl.item()
                decoded_label.append(id2label[lbl])
        decoded_labels.append(decoded_label)

    return decoded_labels


class AutoModelForMultiHeadTokenClassificationConfig(PretrainedConfig):
    model_type="AutoModelForMultiHeadTokenClassification"

    def __init__(self, pretrained_model_name_or_path: Union[Path, str] = "bert-base-multilingual-cased", num_labels: int = 5, **kwargs):
        # don't change the order of the lines below
        super().__init__(num_labels=num_labels, **kwargs)
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.real_num_labels = num_labels


class AutoModelForMultiHeadTokenClassification(PreTrainedModel):
    config_class = AutoModelForMultiHeadTokenClassificationConfig

    """
    This class is a wrapper around the AutoModelForTokenClassification class
    that allows for multi-head binary named entity recognition..
    """

    def __init__(self, config: AutoModelForMultiHeadTokenClassificationConfig, *args, **kwargs):
        super().__init__(config)

        # set up pretrained model
        self.model = AutoModelForTokenClassification.from_pretrained(
            config.pretrained_model_name_or_path,
            num_labels=config.num_labels,
            ignore_mismatched_sizes=True,
            *args, **kwargs
        )

        self.id2label = generate_id_to_label_dict(config.num_labels)


    # adapted from https://github.com/huggingface/transformers/blob/faed2ca46fb163082d154aa234fd5d30682d6bf1/src/transformers/models/bert/modeling_bert.py#L1541
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], SequenceClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, num_labels)`, *optional*):
            Labels for computing the multi-head sequence regression loss.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # Create a dictionary of all potential arguments
        kwargs = {
            'attention_mask': attention_mask,
            'token_type_ids': token_type_ids,
            'position_ids': position_ids,
            'head_mask': head_mask,
            'inputs_embeds': inputs_embeds,
            'output_attentions': output_attentions,
            'output_hidden_states': output_hidden_states,
            'return_dict': return_dict,
        }

        # Get the list of parameter names accepted by the model method
        accepted_params = inspect.signature(self.model.forward).parameters.keys()

        # Filter the arguments to only those accepted by the model method
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in accepted_params}
        outputs: SequenceClassifierOutput = self.model(
            input_ids,
            **filtered_kwargs,
        )

        # logits has shape (1, batch_size, sequence_length, num_labels)
        logits = outputs.logits

        loss = None
        if labels is not None:
            decoded_labels = decode_labels(labels=labels, id2label=self.id2label, num_labels=self.config.real_num_labels)
            decoded_labels = torch.tensor(decoded_labels, dtype=torch.float32).to(logits.device)

            weight = (decoded_labels != -100).to(logits.device).float()
            loss_fct = nn.BCEWithLogitsLoss(weight=weight)
            loss = loss_fct(logits, decoded_labels)

        if not return_dict:
            output = logits + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


# register your config and your model
AutoConfig.register("AutoModelForMultiHeadTokenClassification", AutoModelForMultiHeadTokenClassificationConfig)
AutoModel.register(AutoModelForMultiHeadTokenClassificationConfig, AutoModelForMultiHeadTokenClassification)

if __name__ == "__main__":
    # debug
    config = AutoModelForMultiHeadTokenClassificationConfig(pretrained_model_name_or_path="bert-base-multilingual-cased", num_labels=3)
    model = AutoModelForMultiHeadTokenClassification(config)
    print(model)
    model.save_pretrained("test-workdir/model_save_load_test")
    model_reloaded = AutoModelForMultiHeadTokenClassification.from_pretrained("test-workdir/model_save_load_test")
    print(model_reloaded)
    print("Model successfully saved and reloaded.")
